'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
# XES TAGS
TAG_BOOLEAN = "boolean"
TAG_CLASSIFIER = "classifier"
TAG_DATE = "date"
TAG_EVENT = "event"
TAG_EXTENSION = "extension"
TAG_FLOAT = "float"
TAG_GLOBAL = "global"
TAG_ID = "id"
TAG_INT = "int"
TAG_LIST = "list"
TAG_LOG = "log"
TAG_STRING = "string"
TAG_TRACE = "trace"
TAG_VALUES = "values"

# XES 2.0 tags
TAG_CONTAINER = "container"

# XES/INTERNAL KEYS
KEY_CHILDREN = "children"
KEY_KEY = "key"
KEY_KEYS = "keys"
KEY_NAME = "name"
KEY_PREFIX = "prefix"
KEY_SCOPE = "scope"
KEY_URI = "uri"
KEY_VALUE = "value"

DEFAULT_NAME_KEY = "concept:name"
DEFAULT_INSTANCE_KEY = "concept:instance"
DEFAULT_TIMESTAMP_KEY = "time:timestamp"
DEFAULT_START_TIMESTAMP_KEY = "start_timestamp"
DEFAULT_TRACEID_KEY = "concept:name"
DEFAULT_RESOURCE_KEY = "org:resource"
DEFAULT_TRANSITION_KEY = "lifecycle:transition"
DEFAULT_GROUP_KEY = "org:group"
