'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.algo.discovery.causal.variants import alpha, heuristic
from enum import Enum
from pm4py.util import exec_utils
from typing import Dict, Tuple


class Variants(Enum):
    CAUSAL_ALPHA = alpha
    CAUSAL_HEURISTIC = heuristic


CAUSAL_ALPHA = Variants.CAUSAL_ALPHA
CAUSAL_HEURISTIC = Variants.CAUSAL_HEURISTIC

VERSIONS = {CAUSAL_ALPHA, CAUSAL_HEURISTIC}


def apply(
    dfg: Dict[Tuple[str, str], int], variant=CAUSAL_ALPHA
) -> Dict[Tuple[str, str], int]:
    """
    Computes the causal relation on the basis of a given directly follows graph.

    Parameters
    -----------
    dfg
        Directly follows graph
    variant
        Variant of the algorithm to use:
            - Variants.CAUSAL_ALPHA
            - Variants.CAUSAL_HEURISTIC

    Returns
    -----------
    causal relations
        dict
    """
    return exec_utils.get_variant(variant).apply(dfg)
